"""ISO7816-4 error checking strategy.

__author__ = "https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

import smartcard.sw.SWExceptions
from smartcard.sw.ErrorChecker import ErrorChecker

# pylint: disable=too-few-public-methods

iso7816_4SW = {
    0x62: (
        smartcard.sw.SWExceptions.WarningProcessingException,
        {
            0x00: "Response padded/ More APDU commands expected",
            0x81: "Part of returned data may be corrupted",
            0x82: "End of file/record reached before reading Le bytes",
            0x83: "File invalidated",
            0x84: "FCI not correctly formatted",
            0xFF: "Correct execution, response padded",
        },
    ),
    0x63: (
        smartcard.sw.SWExceptions.WarningProcessingException,
        {
            0x00: "Authentication failed",
            0x81: "File filled up by the last write",
            0xC0: "PIN verification failed. 0 tries before blocking PIN",
            0xC1: "PIN verification failed. 1 tries before blocking PIN",
            0xC2: "PIN verification failed. 2 tries before blocking PIN",
            0xC3: "PIN verification failed. 3 tries before blocking PIN",
            0xC4: "PIN verification failed. 4 tries before blocking PIN",
            0xC5: "PIN verification failed. 5 tries before blocking PIN",
            0xC6: "PIN verification failed. 6 tries before blocking PIN",
            0xC7: "PIN verification failed. 7 tries before blocking PIN",
            0xC8: "PIN verification failed. 8 tries before blocking PIN",
            0xC9: "PIN verification failed. 9 tries before blocking PIN",
            0xCA: "PIN verification failed. 10 tries before blocking PIN",
            0xCB: "PIN verification failed. 11 tries before blocking PIN",
            0xCC: "PIN verification failed. 12 tries before blocking PIN",
            0xCD: "PIN verification failed. 13 tries before blocking PIN",
            0xCE: "PIN verification failed. 14 tries before blocking PIN",
            0xCF: "PIN verification failed. 15 tries before blocking PIN",
        },
    ),
    0x64: (
        smartcard.sw.SWExceptions.ExecutionErrorException,
        {0x00: "Integrity error detected in EEPROM"},
    ),
    0x67: (
        smartcard.sw.SWExceptions.CheckingErrorException,
        {0x00: "Wrong length in Lc"},
    ),
    0x68: (
        smartcard.sw.SWExceptions.CheckingErrorException,
        {0x81: "Logical channel not supported", 0x82: "Secure messaging not supported"},
    ),
    0x69: (
        smartcard.sw.SWExceptions.CheckingErrorException,
        {
            0x81: "Command incompatible with file structure.",
            0x82: "Security status not satisfied",
            0x83: "Authentication method blocked",
            0x84: "Referenced data invalid",
            0x85: "Conditions of use not satisfied",
            0x86: "Command not allowed (no current EF)",
            0x87: "Secure messaging data object missing.",
            0x88: "Secure messaging data object incorrect",
        },
    ),
    0x6A: (
        smartcard.sw.SWExceptions.CheckingErrorException,
        {
            0x80: "Incorrect parameters in the data field",
            0x81: "Function not supported",
            0x82: "File not found",
            0x83: "Record not found",
            0x84: "Not enough memory space in the file",
            0x85: "Lc inconsistent with TLV structure",
            0x86: "Incorrect parameters P1-P2",
            0x87: "Lc is inconsistent with P1-P2",
            0x88: "Referenced data not found",
        },
    ),
    0x6B: (
        smartcard.sw.SWExceptions.CheckingErrorException,
        {0x00: "Incorrect parameters P1-P2"},
    ),
    0x6D: (
        smartcard.sw.SWExceptions.CheckingErrorException,
        {0x00: "Instruction (INS) not supported"},
    ),
    0x6E: (
        smartcard.sw.SWExceptions.CheckingErrorException,
        {0x00: "Class (CLA) not supported"},
    ),
    0x6F: (smartcard.sw.SWExceptions.CheckingErrorException, {0x00: "Fatal error"}),
}


class ISO7816_4ErrorChecker(ErrorChecker):
    """ISO7816-4 error checking strategy.

    This strategy raises the following exceptions:
      - sw1 sw2
      - 62  00 81 82 83 84 FF   WarningProcessingException
      - 63  00 81 C0->CF        WarningProcessingException
      - 64  00                  ExecutionErrorException
      - 67  00                  CheckingErrorException
      - 68  81 82               CheckingErrorException
      - 69  81->88  99? c1?     CheckingErrorException
      - 6a  80->88              CheckingErrorException
      - 6b  00                  CheckingErrorException
      - 6d  00                  CheckingErrorException
      - 6e  00                  CheckingErrorException
      - 6f  00                  CheckingErrorException

    This checker does not raise exceptions on undefined sw1 values, e.g.:
      - sw1 sw2
      - 65  any
      - 66  any
      - 6c  any

    and on undefined sw2 values, e.g.:
      - sw1 sw2
      - 62  80 85
      - 6b  any except 00


    Use another checker in the error checking chain, e.g., the
    ISO7816_4SW1ErrorChecker, to raise exceptions on these undefined
    values.
    """

    def __call__(self, data, sw1, sw2):
        """Called to test data, sw1 and sw2 for error.

        Derived classes must raise a L{smartcard.sw.SWExceptions} upon error.

        @param data:       apdu response data
        @param sw1:        apdu data status words
        @param sw2:        apdu data status words
        """

        # pylint: disable=duplicate-code

        if sw1 in iso7816_4SW:
            exception, sw2dir = iso7816_4SW[sw1]
            if isinstance(sw2dir, dict):
                try:
                    message = sw2dir[sw2]
                    raise exception(data, sw1, sw2, message)
                except KeyError:
                    pass


if __name__ == "__main__":
    # Small sample illustrating the use of ISO7816_4ErrorChecker.
    ecs = ISO7816_4ErrorChecker()
    ecs([], 0x90, 0x00)
    try:
        ecs([], 0x6B, 0x00)
    except smartcard.sw.SWExceptions.CheckingErrorException as e:
        print(str(e) + f" {e.sw1:x} {e.sw2:x}")



"""The CardConnectionDecorator is a Decorator around the CardConnection
abstract class, and allows dynamic addition of features to the
CardConnection, e.g. implementing a secure channel..

__author__ = "https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""


class CardConnectionDecorator:
    """Card connection decorator class."""

    def __init__(self, cardConnectionComponent):
        """Construct a new card connection decorator.

        CardConnectionComponent: CardConnection component to decorate
        """
        self.component = cardConnectionComponent

    def addSWExceptionToFilter(self, exClass):
        """call inner component addSWExceptionToFilter"""
        self.component.addSWExceptionToFilter(exClass)

    def addObserver(self, observer):
        """call inner component addObserver"""
        self.component.addObserver(observer)

    def deleteObserver(self, observer):
        """call inner component deleteObserver"""
        self.component.deleteObserver(observer)

    def connect(self, protocol=None, mode=None, disposition=None):
        """call inner component connect"""
        self.component.connect(protocol, mode, disposition)

    def reconnect(self, protocol=None, mode=None, disposition=None):
        """call inner component reconnect"""
        self.component.reconnect(protocol, mode, disposition)

    def disconnect(self):
        """call inner component disconnect"""
        self.component.disconnect()

    def release(self):
        """call inner component release"""
        self.component.release()

    def getATR(self):
        """call inner component getATR"""
        return self.component.getATR()

    def getProtocol(self):
        """call inner component getProtocol"""
        return self.component.getProtocol()

    def getReader(self):
        """call inner component getReader"""
        return self.component.getReader()

    def setErrorCheckingChain(self, errorcheckingchain):
        """call inner component setErrorCheckingChain"""
        self.component.setErrorCheckingChain(errorcheckingchain)

    def setProtocol(self, protocol):
        """call inner component setProtocol"""
        return self.component.setProtocol(protocol)

    def transmit(self, command, protocol=None):
        """call inner component transmit"""
        return self.component.transmit(command, protocol)

    def control(self, controlCode, command=None):
        """call inner component control"""
        if command is None:
            command = []
        return self.component.control(controlCode, command)

    def getAttrib(self, attribId):
        """call inner component getAttrib"""
        return self.component.getAttrib(attribId)

    def __enter__(self):
        return self

    def __exit__(self, e_type, value, traceback):
        self.disconnect()
        self.release()



"""Card service abstract class.

A card service is a class providings specific smart card functionality,
e.g.  a GSM file system or an Open Platform loader.  CardService is an
abstract class from which concrete card services are derived.  A concrete
card service is almost always smart card operating system specific

__author__ = "https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

from smartcard import CardService


class PassThruCardService(CardService.CardService):
    """Pass-thru card service class."""

    def __init__(self, connection, cardname=None):
        """Construct a pass-thru card service.

        @param connection:     the CardConnection used to access the smart card
        """
        CardService.CardService.__init__(self, connection, cardname)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        return CardService.CardService.__exit__(
            self, exc_type, exc_value, exc_traceback
        )

    @staticmethod
    def supports(cardname):
        """Returns True if the cardname is supported by the card service.
        The PassThruCardService supports all cardnames and always
        returns True."""
        return True


if __name__ == "__main__":
    # Small sample illustrating the use of CardService.
    SELECT = [0xA0, 0xA4, 0x00, 0x00, 0x02]
    DF_TELECOM = [0x7F, 0x10]
    from smartcard.System import readers

    cc = readers()[0].createConnection()
    cs = PassThruCardService(cc)
    cs.connection.connect()
    data, sw1, sw2 = cs.connection.transmit(SELECT + DF_TELECOM)
    print(f"{sw1:X} {sw2:X}")
    cs.connection.disconnect()



"""Smartcard module exceptions.

This module defines the exceptions raised by the smartcard module.

__author__ = "https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

from smartcard.scard import SCardGetErrorMessage


class SmartcardException(Exception):
    """Base class for smartcard exceptions.

    smartcard exceptions are generated by the smartcard module and
    shield scard (i.e. PCSC) exceptions raised by the scard module.

    """

    def __init__(self, *args, message="", hresult=-1):
        if not message and len(args) > 0:
            message = args[0]
            args = args[1:]
        if -1 == hresult and len(args) > 0:
            hresult = args[0]
            args = args[1:]
        super().__init__(message, *args)
        self.hresult = int(hresult)

    def __str__(self):
        text = super().__str__()
        if self.hresult != -1:
            if text:
                text += ": "
            hresult = self.hresult
            if hresult < 0:
                # convert 0x-7FEFFFE3 into 0x8010001D
                hresult += 0x100000000
            text += f"{SCardGetErrorMessage(self.hresult)} (0x{hresult:08X})"

        return text


class CardConnectionException(SmartcardException):
    """Raised when a CardConnection class method fails."""


class CardRequestException(SmartcardException):
    """Raised when a CardRequest wait fails."""


class CardRequestTimeoutException(SmartcardException):
    """Raised when a CardRequest times out."""

    def __init__(self, *args, hresult=-1):
        if -1 == hresult and len(args) > 0:
            hresult = args[0]
            args = args[1:]
        SmartcardException.__init__(
            self, "Time-out during card request", hresult=hresult, *args
        )


class CardServiceException(SmartcardException):
    """Raised when a CardService class method fails."""


class CardServiceStoppedException(SmartcardException):
    """Raised when the CardService was stopped"""


class CardServiceNotFoundException(SmartcardException):
    """Raised when the CardService is not found"""


class InvalidATRMaskLengthException(SmartcardException):
    """Raised when an ATR mask does not match an ATR length."""

    def __init__(self, mask):
        SmartcardException.__init__(self, f"Invalid ATR mask length: {mask}")


class InvalidReaderException(SmartcardException):
    """Raised when trying to access an invalid smartcard reader."""

    def __init__(self, readername):
        SmartcardException.__init__(self, f"Invalid reader: {readername}")


class ListReadersException(SmartcardException):
    """Raised when smartcard readers cannot be listed."""

    def __init__(self, hresult):
        SmartcardException.__init__(self, "Failed to list readers", hresult=hresult)


class NoCardException(SmartcardException):
    """Raised when no card in is present in reader."""

    def __init__(self, message, hresult):
        SmartcardException.__init__(self, message, hresult=hresult)


class NoReadersException(SmartcardException):
    """Raised when the system has no smartcard reader."""

    def __init__(self, *args):
        SmartcardException.__init__(self, "No reader found", *args)



"""CardConnectionObserver interface.

CardConnectionObserver is a base class for objects that are to be notified
upon CardConnection events.

__author__ = "https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

from smartcard.Observer import Observer
from smartcard.util import toHexString

# pylint: disable=too-few-public-methods


# ReaderObserver interface
class CardConnectionObserver(Observer):
    """
    CardConnectionObserver is a base class for objects that are to be notified
    upon L{CardConnection} events.
    """

    def update(self, observable, handlers):
        """Called upon CardConnection event.

        @param observable:         the observed card connection object
        @param handlers:    the CardConnectionEvent sent by the connection
        """


class ConsoleCardConnectionObserver(CardConnectionObserver):
    """CardConnectionObserver output to the console"""

    def update(self, observable, handlers):

        if "connect" == handlers.type:
            print("connecting to " + observable.getReader())

        elif "reconnect" == handlers.type:
            print("reconnecting to " + observable.getReader())

        elif "disconnect" == handlers.type:
            print("disconnecting from " + observable.getReader())

        elif "release" == handlers.type:
            print("release from " + observable.getReader())

        elif "command" == handlers.type:
            print("> " + toHexString(handlers.args[0]))

        elif "response" == handlers.type:
            sw1, sw2 = handlers.args[-2:]
            SW = f" {sw1:2X} {sw2:02X}"
            if [] == handlers.args[0]:
                print("<  []" + SW)
            else:
                print("< " + toHexString(handlers.args[0]) + SW)
        else:
            print("unknown event:", handlers.type)



"""PCSCReader: concrete reader class for PCSC Readers

__author__ = "gemalto https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

from smartcard.CardConnectionDecorator import CardConnectionDecorator
from smartcard.Exceptions import (
    CardServiceNotFoundException,
    CardServiceStoppedException,
    NoCardException,
)
from smartcard.pcsc.PCSCCardConnection import PCSCCardConnection
from smartcard.pcsc.PCSCExceptions import (
    AddReaderToGroupException,
    EstablishContextException,
    IntroduceReaderException,
    ListReadersException,
    ReleaseContextException,
    RemoveReaderFromGroupException,
)
from smartcard.reader.Reader import Reader
from smartcard.scard import (
    SCARD_E_DUPLICATE_READER,
    SCARD_E_NO_READERS_AVAILABLE,
    SCARD_E_NO_SERVICE,
    SCARD_E_SERVICE_STOPPED,
    SCARD_S_SUCCESS,
    SCARD_SCOPE_USER,
    SCardAddReaderToGroup,
    SCardEstablishContext,
    SCardIntroduceReader,
    SCardListReaders,
    SCardReleaseContext,
    SCardRemoveReaderFromGroup,
)


def __PCSCreaders__(hcontext, groups=None):
    """Returns the list of PCSC smartcard readers in PCSC group.

    If group is not specified, returns the list of all PCSC smartcard readers.
    """

    if groups is None:
        groups = []
    elif isinstance(groups, str):
        groups = [groups]
    hresult, readers = SCardListReaders(hcontext, groups)
    if hresult != SCARD_S_SUCCESS:
        if hresult == SCARD_E_NO_READERS_AVAILABLE:
            readers = []
        elif hresult == SCARD_E_SERVICE_STOPPED:
            raise CardServiceStoppedException(hresult=hresult)
        elif hresult == SCARD_E_NO_SERVICE:
            raise CardServiceNotFoundException(hresult=hresult)
        else:
            raise ListReadersException(hresult)

    return readers


class PCSCReader(Reader):
    """PCSC reader class."""

    def __init__(self, readername):
        """Constructs a new PCSC reader."""
        Reader.__init__(self, readername)

    def addtoreadergroup(self, groupname):
        """Add reader to a reader group."""

        hresult, hcontext = SCardEstablishContext(SCARD_SCOPE_USER)
        if SCARD_S_SUCCESS != hresult:
            raise EstablishContextException(hresult)
        try:
            hresult = SCardIntroduceReader(hcontext, self.name, self.name)
            if hresult not in (SCARD_S_SUCCESS, SCARD_E_DUPLICATE_READER):
                raise IntroduceReaderException(hresult, self.name)
            hresult = SCardAddReaderToGroup(hcontext, self.name, groupname)
            if SCARD_S_SUCCESS != hresult:
                raise AddReaderToGroupException(hresult, self.name, groupname)
        finally:
            hresult = SCardReleaseContext(hcontext)
            if SCARD_S_SUCCESS != hresult:
                raise ReleaseContextException(hresult)

    def removefromreadergroup(self, groupname):
        """Remove a reader from a reader group"""

        hresult, hcontext = SCardEstablishContext(SCARD_SCOPE_USER)
        if SCARD_S_SUCCESS != hresult:
            raise EstablishContextException(hresult)
        try:
            hresult = SCardRemoveReaderFromGroup(hcontext, self.name, groupname)
            if SCARD_S_SUCCESS != hresult:
                raise RemoveReaderFromGroupException(hresult, self.name, groupname)
        finally:
            hresult = SCardReleaseContext(hcontext)
            if SCARD_S_SUCCESS != hresult:
                raise ReleaseContextException(hresult)

    def createConnection(self):
        """Return a card connection thru PCSC reader."""
        return CardConnectionDecorator(PCSCCardConnection(self.name))

    class Factory:
        """Factory to create PCSCReader objects"""

        # pylint: disable=too-few-public-methods

        @staticmethod
        def create(readername):
            """Return a PCSCReader object"""
            return PCSCReader(readername)

    @staticmethod
    def readers(groups=None):
        """Return the list of readers"""
        if groups is None:
            groups = []
        creaders = []
        hresult, hcontext = SCardEstablishContext(SCARD_SCOPE_USER)
        if SCARD_S_SUCCESS != hresult:
            raise EstablishContextException(hresult)
        try:
            pcsc_readers = __PCSCreaders__(hcontext, groups)
        finally:
            hresult = SCardReleaseContext(hcontext)
            if SCARD_S_SUCCESS != hresult:
                raise ReleaseContextException(hresult)

        for reader in pcsc_readers:
            creaders.append(PCSCReader.Factory.create(reader))
        return creaders


if __name__ == "__main__":
    from smartcard.util import toHexString

    SELECT = [0xA0, 0xA4, 0x00, 0x00, 0x02]
    DF_TELECOM = [0x7F, 0x10]

    _creaders = PCSCReader.readers()
    for _reader in _creaders:
        try:
            print(_reader.name)
            connection = _reader.createConnection()
            connection.connect()
            print(toHexString(connection.getATR()))
            data, sw1, sw2 = connection.transmit(SELECT + DF_TELECOM)
            print(f"{sw1:02X} {sw2:02X}")
        except NoCardException:
            print("no card in reader")



"""ATR class managing some of the Answer To Reset content.

__author__ = "https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

from __future__ import annotations

import functools
import operator
import warnings

from smartcard.Exceptions import SmartcardException


class ATR:
    """Parse and represent Answer to Reset sequences.

    Answer to Reset sequences are defined in ISO 7816-3, section 8.
    """

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-public-methods

    clockrateconversion: list[int | str] = [
        372,
        372,
        558,
        744,
        1116,
        1488,
        1860,
        "RFU",
        "RFU",
        512,
        768,
        1024,
        1536,
        2048,
        "RFU",
        "RFU",
    ]
    bitratefactor: list[int | str] = [
        "RFU",
        1,
        2,
        4,
        8,
        16,
        32,
        64,
        12,
        20,
        "RFU",
        "RFU",
        "RFU",
        "RFU",
        "RFU",
        "RFU",
    ]
    currenttable: list[int | str] = [25, 50, 100, "RFU"]

    def __init__(self, atr: list[int]) -> None:
        """Parse ATR and initialize members:

          - TS: initial character
          - T0: format character
          - TA[n], TB[n], TC[n], TD[n], for n=0,1,...: protocol parameters

        @note: protocol parameters indices start at 0, e.g.
        TA[0], TA[1] correspond to the ISO standard TA1, TA2
        parameters

          - historicalBytes: the ATR T1, T2, ..., TK historical bytes
          - TCK: checksum byte (only for protocols different from T=0)
          - FI: clock rate conversion factor
          - DI: voltage adjustment factor
          - PI1: programming voltage factor
          - II: maximum programming current factor
          - N: extra guard time
        """

        # pylint: disable=too-many-statements

        if len(atr) < 2:
            raise SmartcardException("ATR sequences must be at least 2 bytes long")
        if atr[0] not in {0x3B, 0x3F}:
            raise SmartcardException(f"invalid TS 0x{atr[0]:02x}")

        self.atr = atr

        # initial character
        self.TS = self.atr[0]

        # format character
        self.T0 = self.atr[1]

        # count of historical bytes
        self.K = self.T0 & 0x0F

        # initialize optional characters lists
        self.TA: list[None | int] = []
        self.TB: list[None | int] = []
        self.TC: list[None | int] = []
        self.TD: list[None | int] = []
        self.Y: list[int] = []

        td: None | int = self.T0
        offset = 1
        while td is not None:
            self.Y.append(td >> 4 & 0x0F)

            self.TA += [None]
            self.TB += [None]
            self.TC += [None]
            self.TD += [None]

            if self.Y[-1] & 0x01:  # TA
                offset += 1
                self.TA[-1] = self.atr[offset]
            if self.Y[-1] & 0x02:  # TB
                offset += 1
                self.TB[-1] = self.atr[offset]
            if self.Y[-1] & 0x04:  # TC
                offset += 1
                self.TC[-1] = self.atr[offset]
            if self.Y[-1] & 0x08:  # TD
                offset += 1
                self.TD[-1] = self.atr[offset]

            td = self.TD[-1]

        self.interfaceBytesCount = offset - 1

        # historical bytes
        self.historicalBytes = self.atr[offset + 1 : offset + 1 + self.K]

        # checksum
        self.TCK: int | None = None
        self.checksumOK: bool | None = None
        self.hasChecksum = len(self.atr) == offset + 1 + self.K + 1
        if self.hasChecksum:
            self.TCK = self.atr[-1]
            self.checksumOK = functools.reduce(operator.xor, self.atr[1:]) == 0

        # clock-rate conversion factor
        self.FI: int | None = None
        if self.TA[0] is not None:
            self.FI = self.TA[0] >> 4 & 0x0F

        # bit-rate adjustment factor
        self.DI: int | None = None
        if self.TA[0] is not None:
            self.DI = self.TA[0] & 0x0F

        # maximum programming current factor
        self.II: int | None = None
        if self.TB[0] is not None:
            self.II = self.TB[0] >> 5 & 0x03

        # programming voltage factor
        self.PI1: int | None = None
        if self.TB[0] is not None:
            self.PI1 = self.TB[0] & 0x1F

        # extra guard time
        self.N = self.TC[0]

    @property
    def hasTA(self) -> list[bool]:
        """Deprecated. Replace usage with `ATR.TA[i] is not None`."""

        warnings.warn("Replace usage with `ATR.TA[i] is not None`", DeprecationWarning)
        return [ta is not None for ta in self.TA]

    @property
    def hasTB(self) -> list[bool]:
        """Deprecated. Replace usage with `ATR.TB[i] is not None`."""

        warnings.warn("Replace usage with `ATR.TB[i] is not None`", DeprecationWarning)
        return [tb is not None for tb in self.TB]

    @property
    def hasTC(self) -> list[bool]:
        """Deprecated. Replace usage with `ATR.TC[i] is not None`."""

        warnings.warn("Replace usage with `ATR.TC[i] is not None`", DeprecationWarning)
        return [tc is not None for tc in self.TC]

    @property
    def hasTD(self) -> list[bool]:
        """Deprecated. Replace usage with `ATR.TD[i] is not None`."""

        warnings.warn("Replace usage with `ATR.TD[i] is not None`", DeprecationWarning)
        return [td is not None for td in self.TD]

    def getChecksum(self) -> int | None:
        """Return the checksum of the ATR. Checksum is mandatory only
        for T=1."""
        return self.TCK

    def getHistoricalBytes(self) -> list[int]:
        """Return historical bytes."""
        return self.historicalBytes

    def getHistoricalBytesCount(self) -> int:
        """Return count of historical bytes."""
        return len(self.historicalBytes)

    def getInterfaceBytesCount(self) -> int:
        """Return count of interface bytes."""
        return self.interfaceBytesCount

    def getTA1(self) -> int | None:
        """Return TA1 byte."""
        return self.TA[0]

    def getTB1(self) -> int | None:
        """Return TB1 byte."""
        return self.TB[0]

    def getTC1(self) -> int | None:
        """Return TC1 byte."""
        return self.TC[0]

    def getTD1(self) -> int | None:
        """Return TD1 byte."""
        return self.TD[0]

    def getBitRateFactor(self) -> int | str:
        """Return bit rate factor."""
        if self.DI is not None:
            return ATR.bitratefactor[self.DI]
        return 1

    def getClockRateConversion(self) -> int | str:
        """Return clock rate conversion."""
        if self.FI is not None:
            return ATR.clockrateconversion[self.FI]
        return 372

    def getProgrammingCurrent(self) -> int | str:
        """Return maximum programming current."""
        if self.II is not None:
            return ATR.currenttable[self.II]
        return 50

    def getProgrammingVoltage(self) -> int:
        """Return programming voltage."""
        if self.PI1 is not None:
            return 5 * (1 + self.PI1)
        return 5

    def getGuardTime(self) -> int | None:
        """Return extra guard time."""
        return self.N

    def getSupportedProtocols(self) -> dict[str, bool]:
        """Returns a dictionary of supported protocols."""
        protocols: dict[str, bool] = {}
        for td in self.TD:
            if td is not None:
                protocols[f"T={td & 0x0F}"] = True
        if self.TD[0] is None:
            protocols["T=0"] = True
        return protocols

    def isT0Supported(self) -> bool:
        """Return True if T=0 is supported."""
        return "T=0" in self.getSupportedProtocols()

    def isT1Supported(self) -> bool:
        """Return True if T=1 is supported."""
        return "T=1" in self.getSupportedProtocols()

    def isT15Supported(self) -> bool:
        """Return True if T=15 is supported."""
        return "T=15" in self.getSupportedProtocols()

    def render(self) -> str:
        """Render the ATR to a readable format."""

        lines: list[str] = []
        enumerated_tx_values = enumerate(zip(self.TA, self.TB, self.TC, self.TD), 1)
        for i, (ta, tb, tc, td) in enumerated_tx_values:
            if ta is not None:
                lines.append(f"TA{i}: {ta:x}")
            if tb is not None:
                lines.append(f"TB{i}: {tb:x}")
            if tc is not None:
                lines.append(f"TC{i}: {tc:x}")
            if td is not None:
                lines.append(f"TD{i}: {td:x}")

        lines.append(f"supported protocols {','.join(self.getSupportedProtocols())}")
        lines.append(f"T=0 supported: {self.isT0Supported()}")
        lines.append(f"T=1 supported: {self.isT1Supported()}")

        if self.getChecksum() is not None:
            lines.append(f"checksum: {self.getChecksum()}")

        lines.append(f"\tclock rate conversion factor: {self.getClockRateConversion()}")
        lines.append(f"\tbit rate adjustment factor: {self.getBitRateFactor()}")
        lines.append(f"\tmaximum programming current: {self.getProgrammingCurrent()}")
        lines.append(f"\tprogramming voltage: {self.getProgrammingVoltage()}")
        lines.append(f"\tguard time: {self.getGuardTime()}")
        lines.append(f"nb of interface bytes: {self.getInterfaceBytesCount()}")
        lines.append(f"nb of historical bytes: {self.getHistoricalBytesCount()}")

        return "\n".join(lines)

    def dump(self) -> None:
        """Deprecated. Replace usage with `print(ATR.render())`"""

        warnings.warn("Replace usage with `print(ATR.render())`", DeprecationWarning)
        print(self.render())

    def __str__(self) -> str:
        """Render the ATR as a space-separated string of uppercase hexadecimal pairs."""

        return bytes(self.atr).hex(" ").upper()



"""The CardConnection abstract class manages connections with a card and
apdu transmission.

__author__ = "https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

from smartcard.CardConnectionEvent import CardConnectionEvent
from smartcard.Observer import Observable


class CardConnection(Observable):
    """Card connection abstract class."""

    T0_protocol = 0x00000001
    """ protocol T=0 """

    T1_protocol = 0x00000002
    """ protocol T=1 """

    RAW_protocol = 0x00010000
    """ protocol RAW (direct access to the reader) """

    T15_protocol = 0x00000008
    """ protocol T=15 """

    def __init__(self, reader):
        """Construct a new card connection.

        @param reader: name of the reader in which the smartcard to connect
        to is located.
        """
        Observable.__init__(self)
        self.reader = reader
        """ reader name """
        self.errorcheckingchain = None
        """ see L{setErrorCheckingChain} """
        self.defaultprotocol = CardConnection.T0_protocol | CardConnection.T1_protocol
        """ see L{setProtocol} and L{getProtocol} """

    def __del__(self):
        """Connect to card."""

    def addSWExceptionToFilter(self, exClass):
        """Add a status word exception class to be filtered.

        @param exClass: the class to filter, e.g.
        L{smartcard.sw.SWExceptions.WarningProcessingException}

        Filtered exceptions will not be raised when encountered in the
        error checking chain."""
        if self.errorcheckingchain is not None:
            self.errorcheckingchain[0].addFilterException(exClass)

    def addObserver(self, observer):
        """Add a L{CardConnection} observer."""
        Observable.addObserver(self, observer)

    def deleteObserver(self, observer):
        """Remove a L{CardConnection} observer."""
        Observable.deleteObserver(self, observer)

    def connect(self, protocol=None, mode=None, disposition=None):
        """Connect to card.
        @param protocol: a bit mask of the protocols to use, from
        L{CardConnection.T0_protocol}, L{CardConnection.T1_protocol},
        L{CardConnection.RAW_protocol}, L{CardConnection.T15_protocol}

        @param mode: C{smartcard.scard.SCARD_SHARE_SHARED} (default),
        C{smartcard.scard.SCARD_SHARE_EXCLUSIVE} or
        C{smartcard.scard.SCARD_SHARE_DIRECT}

        @param disposition: C{smartcard.scard.SCARD_LEAVE_CARD}
        (default), C{smartcard.scard.SCARD_RESET_CARD},
        C{smartcard.scard.SCARD_UNPOWER_CARD} or
        C{smartcard.scard.SCARD_EJECT_CARD}
        """
        # pylint: disable=unused-argument
        Observable.setChanged(self)
        Observable.notifyObservers(self, CardConnectionEvent("connect"))

    def reconnect(self, protocol=None, mode=None, disposition=None):
        """Reconnect to card.
        @param protocol: a bit mask of the protocols to use, from
        L{CardConnection.T0_protocol}, L{CardConnection.T1_protocol},
        L{CardConnection.RAW_protocol}, L{CardConnection.T15_protocol}

        @param mode: C{smartcard.scard.SCARD_SHARE_SHARED} (default),
        C{smartcard.scard.SCARD_SHARE_EXCLUSIVE} or
        C{smartcard.scard.SCARD_SHARE_DIRECT}

        @param disposition: C{smartcard.scard.SCARD_LEAVE_CARD},
        C{smartcard.scard.SCARD_RESET_CARD} (default),
        C{smartcard.scard.SCARD_UNPOWER_CARD} or
        C{smartcard.scard.SCARD_EJECT_CARD}
        """
        # pylint: disable=unused-argument
        Observable.setChanged(self)
        Observable.notifyObservers(self, CardConnectionEvent("reconnect"))

    def disconnect(self):
        """Disconnect from card."""
        Observable.setChanged(self)
        Observable.notifyObservers(self, CardConnectionEvent("disconnect"))

    def release(self):
        """Release the context."""
        Observable.setChanged(self)
        Observable.notifyObservers(self, CardConnectionEvent("release"))

    def getATR(self):
        """Return card ATR"""

    def getProtocol(self):
        """Return bit mask for the protocol of connection, or None if no
        protocol set.  The return value is a bit mask of
        L{CardConnection.T0_protocol}, L{CardConnection.T1_protocol},
        L{CardConnection.RAW_protocol}, L{CardConnection.T15_protocol}
        """
        return self.defaultprotocol

    def getReader(self):
        """Return card connection reader"""
        return self.reader

    def setErrorCheckingChain(self, errorcheckingchain):
        """Add an error checking chain.
        @param errorcheckingchain: a L{smartcard.sw.ErrorCheckingChain}
        object The error checking strategies in errorchecking chain will
        be tested with each received response APDU, and a
        L{smartcard.sw.SWExceptions.SWException} will be raised upon
        error."""
        self.errorcheckingchain = errorcheckingchain

    def setProtocol(self, protocol):
        """Set protocol for card connection.
        @param protocol: a bit mask of L{CardConnection.T0_protocol},
        L{CardConnection.T1_protocol}, L{CardConnection.RAW_protocol},
        L{CardConnection.T15_protocol}

        >>> setProtocol(CardConnection.T1_protocol | CardConnection.T0_protocol)
        """
        self.defaultprotocol = protocol

    def transmit(self, command, protocol=None):
        """Transmit an apdu. Internally calls L{doTransmit()} class method
        and notify observers upon command/response APDU events.
        Subclasses must override the L{doTransmit()} class method.

        @param command:    list of bytes to transmit

        @param protocol:   the transmission protocol, from
                    L{CardConnection.T0_protocol},
                    L{CardConnection.T1_protocol}, or
                    L{CardConnection.RAW_protocol}
        """
        Observable.setChanged(self)
        Observable.notifyObservers(
            self, CardConnectionEvent("command", [command, protocol])
        )
        data, sw1, sw2 = self.doTransmit(command, protocol)
        Observable.setChanged(self)
        Observable.notifyObservers(
            self, CardConnectionEvent("response", [data, sw1, sw2])
        )
        if self.errorcheckingchain is not None:
            self.errorcheckingchain[0](data, sw1, sw2)
        return data, sw1, sw2

    def doTransmit(self, command, protocol):
        """Performs the command APDU transmission.

        Subclasses must override this method for implementing apdu
        transmission."""
        # pylint: disable=unused-argument
        return [], 0, 0

    def control(self, controlCode, command=None):
        """Send a control command and buffer.  Internally calls
        L{doControl()} class method and notify observers upon
        command/response events.  Subclasses must override the
        L{doControl()} class method.

        @param controlCode: command code

        @param command:     list of bytes to transmit
        """
        if command is None:
            command = []
        Observable.setChanged(self)
        Observable.notifyObservers(
            self, CardConnectionEvent("command", [controlCode, command])
        )
        data = self.doControl(controlCode, command)
        Observable.setChanged(self)
        Observable.notifyObservers(self, CardConnectionEvent("response", data))
        if self.errorcheckingchain is not None:
            self.errorcheckingchain[0](data)
        return data

    def doControl(self, controlCode, command):
        """Performs the command control.

        Subclasses must override this method for implementing control."""
        # pylint: disable=unused-argument
        return []

    def getAttrib(self, attribId):
        """return the requested attribute

        @param attribId: attribute id like
        C{smartcard.scard.SCARD_ATTR_VENDOR_NAME}
        """
        Observable.setChanged(self)
        Observable.notifyObservers(self, CardConnectionEvent("attrib", [attribId]))
        data = self.doGetAttrib(attribId)
        if self.errorcheckingchain is not None:
            self.errorcheckingchain[0](data)
        return data

    def doGetAttrib(self, attribId):
        """Performs the command get attrib.

        Subclasses must override this method for implementing get attrib."""
        # pylint: disable=unused-argument
        return []

    def __enter__(self):
        """Enter the runtime context."""
        return self

    def __exit__(self, e_type, value, traceback):
        """Exit the runtime context trying to disconnect."""
        self.disconnect()



"""
A wxValidator that matches APDU in hexadecimal such as::

    A4 A0 00 00 02
    A4A0000002

__author__ = "https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

import re
import string

import wx

# a regexp to match ATRs and APDUs
hexbyte = "[0-9a-fA-F]{1,2}"
apduregexp = re.compile("((%s)[ ]*)*" % hexbyte)


class APDUHexValidator(wx.PyValidator):
    """A wxValidator that matches APDU in hexadecimal such as:
    A4 A0 00 00 02
    A4A0000002"""

    def __init__(self):
        wx.Validator.__init__(self)
        self.Bind(wx.EVT_CHAR, self.OnChar)

    def Clone(self):
        return APDUHexValidator()

    def Validate(self, win):
        tc = self.GetWindow()
        value = tc.GetValue()

        if not apduregexp.match(value):
            return False

        return True

    def OnChar(self, event):
        key = event.GetKeyCode()

        if wx.WXK_SPACE == key or chr(key) in string.hexdigits:
            value = event.GetEventObject().GetValue() + chr(key)
            if apduregexp.match(value):
                event.Skip()
            return

        if key < wx.WXK_SPACE or key == wx.WXK_DELETE or key > 255:
            event.Skip()
            return

        if not wx.Validator.IsSilent():
            wx.Bell()

        return



"""PCSC Smartcard request.

__author__ = "https://www.gemalto.com/"

Copyright 2001-2012 gemalto
Author: Jean-Daniel Aussel, mailto:jean-daniel.aussel@gemalto.com

This file is part of pyscard.

pyscard is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

pyscard is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with pyscard; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

import threading
from datetime import datetime

from smartcard import Card
from smartcard.AbstractCardRequest import AbstractCardRequest
from smartcard.Exceptions import (
    CardConnectionException,
    CardRequestException,
    CardRequestTimeoutException,
    ListReadersException,
)
from smartcard.pcsc.PCSCReader import PCSCReader
from smartcard.scard import (
    INFINITE,
    SCARD_E_CANCELLED,
    SCARD_E_NO_READERS_AVAILABLE,
    SCARD_E_NO_SERVICE,
    SCARD_E_SERVICE_STOPPED,
    SCARD_E_SYSTEM_CANCELLED,
    SCARD_E_TIMEOUT,
    SCARD_E_UNKNOWN_READER,
    SCARD_S_SUCCESS,
    SCARD_SCOPE_USER,
    SCARD_STATE_CHANGED,
    SCARD_STATE_PRESENT,
    SCARD_STATE_UNAWARE,
    SCardCancel,
    SCardEstablishContext,
    SCardGetErrorMessage,
    SCardGetStatusChange,
    SCardListReaders,
    SCardReleaseContext,
)


class PCSCCardRequest(AbstractCardRequest):
    """PCSC CardRequest class."""

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-positional-arguments
    def __init__(
        self,
        newcardonly=False,
        readers=None,
        cardType=None,
        cardServiceClass=None,
        timeout=1,
    ):
        """Construct new PCSCCardRequest.

        @param newcardonly: if C{True}, request a new card. default is
        C{False}, i.e. accepts cards already inserted

        @param readers: the list of readers to consider for requesting a
        card default is to consider all readers

        @param cardType: the L{CardType} class to wait for; default is
        L{AnyCardType}, i.e.  the request will returns with new or already
        inserted cards

        @param cardServiceClass: the specific card service class to
        create and bind to the card default is to create and bind a
        L{PassThruCardService}

        @param timeout: the time in seconds we are ready to wait for
        connecting to the requested card.  default is to wait one second
        to wait forever, set timeout to C{None}
        """
        AbstractCardRequest.__init__(
            self, newcardonly, readers, cardType, cardServiceClass, timeout
        )

        # if timeout is None, translate to scard.INFINITE
        if self.timeout is None:
            self.timeout = INFINITE
        # otherwise, from seconds to milliseconds
        else:
            self.timeout = int(self.timeout * 1000)

        hresult, self.hcontext = SCardEstablishContext(SCARD_SCOPE_USER)
        if hresult != SCARD_S_SUCCESS:
            self.hcontext = None
            raise CardConnectionException(hresult=hresult)
        self.evt = threading.Event()
        self.hresult = SCARD_S_SUCCESS
        self.readerstates = {}
        self.newstates = []
        self.timeout_init = self.timeout

    def __del__(self):
        self.release()

    def release(self):
        """Release the PCSC context"""
        if self.hcontext is not None:
            hresult = SCardReleaseContext(self.hcontext)
            if hresult != SCARD_S_SUCCESS:
                raise CardConnectionException(hresult=hresult)
            self.hcontext = None

    def getReaderNames(self):
        """Returns the list of PCSC readers on which to wait for cards."""

        # get inserted readers
        hresult, pcscreaders = SCardListReaders(self.hcontext, [])

        # renew the context in case PC/SC was stopped
        # this happens on Windows when the last reader is disconnected
        if hresult in (SCARD_E_SERVICE_STOPPED, SCARD_E_NO_SERVICE):
            hresult = SCardReleaseContext(self.hcontext)
            if hresult != SCARD_S_SUCCESS:
                raise CardConnectionException(hresult=hresult)
            hresult, self.hcontext = SCardEstablishContext(SCARD_SCOPE_USER)
            if hresult != SCARD_S_SUCCESS:
                raise CardConnectionException(hresult=hresult)
            hresult, pcscreaders = SCardListReaders(self.hcontext, [])
        if SCARD_E_NO_READERS_AVAILABLE == hresult:
            return []
        if SCARD_S_SUCCESS != hresult:
            raise ListReadersException(hresult)

        readers = []

        # if no readers asked, use all inserted readers
        if self.readersAsked is None:
            readers = pcscreaders

        # otherwise use only the asked readers that are inserted
        else:
            for reader in self.readersAsked:
                if not isinstance(reader, str):
                    reader = str(reader)
                if reader in pcscreaders:
                    readers = readers + [reader]

        return readers

    # thread waiting for a change
    # the main thread will handle a possible KeyboardInterrupt
    def __getStatusChange(self):
        self.hresult, self.newstates = SCardGetStatusChange(
            self.hcontext, self.timeout, list(self.readerstates.values())
        )
        self.evt.set()

    def waitforcard(self):
        """Wait for card insertion and returns a card service."""

        # pylint: disable=too-many-statements
        # pylint: disable=too-many-locals
        # pylint: disable=too-many-branches

        AbstractCardRequest.waitforcard(self)
        cardfound = False

        # create a dictionary entry for new readers
        readerstates = {}
        readernames = self.getReaderNames()
        # add PnP special reader
        readernames.append("\\\\?PnP?\\Notification")

        for reader in readernames:
            if reader not in readerstates:
                readerstates[reader] = (reader, SCARD_STATE_UNAWARE)

        # call SCardGetStatusChange only if we have some readers
        if readerstates:
            hresult, newstates = SCardGetStatusChange(
                self.hcontext, 0, list(readerstates.values())
            )
        else:
            hresult = SCARD_S_SUCCESS
            newstates = []

        # we can expect normally time-outs or reader
        # disappearing just before the call
        # otherwise, raise exception on error
        if hresult not in (SCARD_S_SUCCESS, SCARD_E_TIMEOUT, SCARD_E_UNKNOWN_READER):
            raise CardRequestException(
                "Failed to SCardGetStatusChange " + SCardGetErrorMessage(hresult),
                hresult=hresult,
            )

        # update readerstate
        for state in newstates:
            readername, eventstate, atr = state
            readerstates[readername] = (readername, eventstate)

        # if a new card is not requested, just return the first available
        if not self.newcardonly:
            for state in newstates:
                readername, eventstate, atr = state
                if eventstate & SCARD_STATE_PRESENT:
                    reader = PCSCReader(readername)
                    if self.cardType.matches(atr, reader):
                        if self.cardServiceClass.supports("dummy"):
                            cardfound = True
                            return self.cardServiceClass(reader.createConnection())

        startDate = datetime.now()
        self.timeout = self.timeout_init
        while not cardfound:

            # create a dictionary entry for new readers
            readernames = self.getReaderNames()

            if self.readersAsked is None:
                # add PnP special reader
                readernames.append("\\\\?PnP?\\Notification")

            for reader in readernames:
                if reader not in readerstates:
                    readerstates[reader] = (reader, SCARD_STATE_UNAWARE)

            # remove dictionary entry for readers that disappeared
            for oldreader in list(readerstates.keys()):
                if oldreader not in readernames:
                    del readerstates[oldreader]

            # wait for card insertion
            self.readerstates = readerstates
            waitThread = threading.Thread(target=self.__getStatusChange)
            waitThread.start()

            # the main thread handles a possible KeyboardInterrupt
            try:
                waitThread.join()
            except KeyboardInterrupt as exc:
                hresult = SCardCancel(self.hcontext)
                if hresult != SCARD_S_SUCCESS:
                    raise CardRequestException(
                        "Failed to SCardCancel " + SCardGetErrorMessage(hresult),
                        hresult=hresult,
                    ) from exc

            # wait for the thread to finish in case of KeyboardInterrupt
            self.evt.wait(timeout=None)

            # get values set in the __getStatusChange thread
            hresult = self.hresult
            newstates = self.newstates

            # compute remaining timeout
            if self.timeout != INFINITE:
                delta = datetime.now() - startDate
                self.timeout -= int(delta.total_seconds() * 1000)
                # timeout cant be < 0
                self.timeout = max(self.timeout, 0)

            # time-out
            if hresult in (SCARD_E_TIMEOUT, SCARD_E_CANCELLED):
                raise CardRequestTimeoutException(hresult=hresult)

            # reader vanished before or during the call
            if SCARD_E_UNKNOWN_READER == hresult:
                pass

            # this happens on Windows when the last reader is disconnected
            elif hresult in (SCARD_E_SYSTEM_CANCELLED, SCARD_E_NO_SERVICE):
                pass

            # some error happened
            elif SCARD_S_SUCCESS != hresult:
                raise CardRequestException(
                    "Failed to get status change " + SCardGetErrorMessage(hresult),
                    hresult=hresult,
                )

            # something changed!
            else:

                # check if we have to return a match, i.e.
                # if no new card in inserted and there is a card found
                # or if a new card is requested, and there is a change+present
                for state in newstates:
                    readername, eventstate, atr = state
                    _, oldstate = readerstates[readername]

                    # the status can change on a card already inserted, e.g.
                    # unpowered, in use, ...
                    # if a new card is requested, clear the state changed bit
                    # if the card was already inserted and is still inserted
                    if self.newcardonly:
                        if oldstate & SCARD_STATE_PRESENT and eventstate & (
                            SCARD_STATE_CHANGED | SCARD_STATE_PRESENT
                        ):
                            eventstate = eventstate & (0xFFFFFFFF ^ SCARD_STATE_CHANGED)

                    if (
                        self.newcardonly
                        and eventstate & SCARD_STATE_PRESENT
                        and eventstate & SCARD_STATE_CHANGED
                    ) or (not self.newcardonly and eventstate & SCARD_STATE_PRESENT):
                        reader = PCSCReader(readername)
                        if self.cardType.matches(atr, reader):
                            if self.cardServiceClass.supports("dummy"):
                                cardfound = True
                                return self.cardServiceClass(reader.createConnection())

                    # update state dictionary
                    readerstates[readername] = (readername, eventstate)

        # should not happen
        return None

    def waitforcardevent(self):
        """Wait for card insertion or removal."""

        # pylint: disable=too-many-statements
        # pylint: disable=too-many-locals
        # pylint: disable=too-many-branches

        AbstractCardRequest.waitforcardevent(self)
        presentcards = []

        startDate = datetime.now()
        eventfound = False
        self.timeout = self.timeout_init
        while not eventfound:

            # get states from previous run
            readerstates = self.readerstates

            # reinitialize at each iteration just in case a new reader appeared
            _readernames = self.getReaderNames()
            readernames = _readernames

            if self.readersAsked is None:
                # add PnP special reader
                readernames.append("\\\\?PnP?\\Notification")

            # first call?
            if len(readerstates) == 0:
                # init
                for reader in readernames:
                    # create a dictionary entry for new readers
                    readerstates[reader] = (reader, SCARD_STATE_UNAWARE)

            # check if a new reader with a card has just been connected
            for reader in _readernames:
                # is the reader a new one?
                if reader not in readerstates:
                    # create a dictionary entry for new reader
                    readerstates[reader] = (reader, SCARD_STATE_UNAWARE)

                    hresult, newstates = SCardGetStatusChange(
                        self.hcontext, 0, list(readerstates.values())
                    )

                    # added reader is the last one (index is -1)
                    _, state, _ = newstates[-1]
                    if state & SCARD_STATE_PRESENT:
                        eventfound = True

            # check if a reader has been removed
            to_remove = []
            for reader in readerstates:
                if reader not in readernames:
                    _, state = readerstates[reader]
                    # was the card present?
                    if state & SCARD_STATE_PRESENT:
                        eventfound = True

                    to_remove.append(reader)

            if to_remove:
                for reader in to_remove:
                    # remove reader
                    del readerstates[reader]

                # get newstates with new reader list
                hresult, newstates = SCardGetStatusChange(
                    self.hcontext, 0, list(readerstates.values())
                )

            if eventfound:
                break

            # wait for card insertion
            self.readerstates = readerstates
            waitThread = threading.Thread(target=self.__getStatusChange)
            waitThread.start()

            # the main thread handles a possible KeyboardInterrupt
            try:
                waitThread.join()
            except KeyboardInterrupt as exc:
                hresult = SCardCancel(self.hcontext)
                if hresult != SCARD_S_SUCCESS:
                    raise CardRequestException(
                        "Failed to SCardCancel " + SCardGetErrorMessage(hresult),
                        hresult=hresult,
                    ) from exc

            # wait for the thread to finish in case of KeyboardInterrupt
            self.evt.wait(timeout=None)

            # get values set in the __getStatusChange thread
            hresult = self.hresult
            newstates = self.newstates

            # compute remaining timeout
            if self.timeout != INFINITE:
                delta = datetime.now() - startDate
                self.timeout -= int(delta.total_seconds() * 1000)
                # timeout cant be < 0
                self.timeout = max(self.timeout, 0)

            # time-out
            if hresult in (SCARD_E_TIMEOUT, SCARD_E_CANCELLED):
                raise CardRequestTimeoutException(hresult=hresult)

            # the reader was unplugged during the loop
            if SCARD_E_UNKNOWN_READER == hresult:
                pass

            # this happens on Windows when the last reader is disconnected
            elif hresult in (SCARD_E_SYSTEM_CANCELLED, SCARD_E_NO_SERVICE):
                pass

            # some error happened
            elif SCARD_S_SUCCESS != hresult:
                raise CardRequestException(
                    "Failed to get status change " + SCardGetErrorMessage(hresult),
                    hresult=hresult,
                )

            # something changed!
            else:
                for state in newstates:
                    readername, eventstate, atr = state

                    # ignore PnP reader
                    if readername == "\\\\?PnP?\\Notification":
                        continue

                    if eventstate & SCARD_STATE_CHANGED:
                        eventfound = True

            # update readerstates for next SCardGetStatusChange() call
            self.readerstates = {}
            for reader, state, atr in newstates:
                self.readerstates[reader] = (reader, state)

        # return all the cards present
        for state in newstates:
            readername, eventstate, atr = state
            if readername == "\\\\?PnP?\\Notification":
                continue
            if eventstate & SCARD_STATE_PRESENT:
                presentcards.append(Card.Card(readername, atr))

        return presentcards


if __name__ == "__main__":
    # Small sample illustrating the use of PCSCCardRequest.py.

    from smartcard.util import toHexString

    print("Insert a new card within 10 seconds")
    cr = PCSCCardRequest(timeout=10, newcardonly=True)
    cs = cr.waitforcard()
    cs.connection.connect()
    print(cs.connection.getReader() + " " + toHexString(cs.connection.getATR()))
    cs.connection.disconnect()
